
# Term
stcsr_p <- function(p, a, n, b, m, i, j) {
    return(choose(n, i) * choose(m, j) * p**(i + j) *
        (1.0 - p)**(n + m - i - j))
}

# Term, first derivate
stcsr_dp <- function(p, a, n, b, m, i, j) {
    nmij <- n + m - i - j
    return (choose(n, i) * choose(m, j) *
            ((i + j) * p**(i + j - 1.0) * (1.0 - p)**nmij -
              p**(i + j) * nmij * (1.0 - p)**(nmij - 1.0)))
}

# Term, second derivate
stcsr_ddp <- function(p, a, n, b, m, i, j) {
    nmij <- n + m - i - j
    return (choose(n, i) * choose(m, j) *
        ((i + j) * ((i + j - 1.0) * p**(i + j - 2.0) * (1.0 - p)**nmij -
           p**(i + j - 1.0) * nmij * (1.0 - p)**(nmij - 1.0)) -
         nmij * ((i + j) * p**(i + j - 1.0) * (1.0 - p)**(nmij - 1.0) -
           p**(i + j) * (nmij - 1.0) * (1.0 - p)**(nmij - 2.0))))
}

# Main function
mn_stcsr <- function(p, a, n, b, m, func, twotailed) {

    # Statistic
    ST <- 0.0;

    for (i in 0:n) {
        for (j in 0:m) {
            if (((i >= a) && (j <= b)) ||
                    (twotailed && (i <= n - a && j >= m - b))) {
                ST <- ST + func(p, a, n, b, m, i, j);
            }
        }
    }
    return(ST)
}

# S(P)
stcsr_S <- function(p, a, n, b, m) {
    return(mn_stcsr(p, a, n, b, m, stcsr_p, FALSE))
}

# S'(P)
stcsr_dS <- function(p, a, n, b, m) {
    return (mn_stcsr(p, a, n, b, m, stcsr_dp, FALSE))
}

# S''(P)
stcsr_ddS <- function(p, a, n, b, m) {
    return(mn_stcsr(p, a, n, b, m, stcsr_ddp, FALSE))
}

# T(P)
stcsr_T <- function(p, a, n, b, m) {
    return(mn_stcsr(p, a, n, b, m, stcsr_p, TRUE))
}

# T'(P)
stcsr_dT <- function(p, a, n, b, m) {
    return(mn_stcsr(p, a, n, b, m, stcsr_dp, TRUE))
}

# T''(P)
stcsr_ddT <- function(p, a, n, b, m) {
    return(mn_stcsr(p, a, n, b, m, stcsr_ddp, TRUE))
}

# Newton-Raphson
newton_raphson <- function(a, n, b, m, f, df, ddf) {

    p <- (a + b) / (n + m)
    EPSILON <- 1e-03

    repeat {

        p_1 <- p

        p <- p - df(p, a, n, b, m) / ddf(p, a, n, b, m)

        # Check convergence
        if (f(p, a, n, b, m) - f(p_1, a, n, b, m) < 0.0)
            return(p_1)

        if (abs(p - p_1) < EPSILON) {
            return(p_1)
        }

    }

    return(p)
}

# Taillard Statistical Test
taillard.test <- function(x, alternative = "two.sided") {

    DNAME <- deparse(substitute(x))

    # Check parameters
    if (!is.matrix(x) || any(dim(x) != 2)) {
        stop("'x' must be a 2x2 matrix")
    }

    if (!is.numeric(x) || any(x < 0) || any(is.na(x))) {
        stop("all entries of 'x' must be nonnegative and finite")
    }

    if (!is.integer(x)) {
        xo <- x
        x <- round(x)

        if(!identical(TRUE, (ax <- all.equal(xo, x))))
            warning("'x' has been rounded to integer: ", ax)

        storage.mode(x) <- "integer"
    }

    alternative <- char.expand(alternative, c("two.sided", "one.sided"))
    if (length(alternative) > 1 || is.na(alternative)) {
        stop("alternative must be \"two.sided\" or \"one.sided\"")
    }

    # Compute 'p' that maximize function
    RVAL <- NULL
    n <- x[1] + x[2]
    m <- x[3] + x[4]
    switch (alternative,
        one.sided = {
            p <- newton_raphson(x[1], n, x[3], m, stcsr_S, stcsr_dS,
                                stcsr_ddS)
            p_value <- stcsr_S(p, x[1], n, x[3], m)
        },
        two.sided = {
            p <- newton_raphson(x[1], n, x[3], m, stcsr_T, stcsr_dT,
                                stcsr_ddT)
            p_value <- stcsr_T(p, x[1], n, x[3], m)
        }
    )

    RVAL <- list(p.value = p_value,
              alternative = alternative,
              method = "Taillard Test to Compare Proportions",
              data.name = DNAME)

    attr(RVAL, "class") <- "htest"

    return (RVAL)
}


